/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/add_keyset.hh"
#include "src/exception.hh"

#include <libfred/object/check_handle.hh>
#include <libfred/poll/create_update_object_poll_message.hh>
#include <libfred/registrable_object/domain/update_domain.hh>
#include <libfred/registrable_object/keyset/create_keyset.hh>
#include <libfred/registrable_object/keyset/info_keyset.hh>
#include <util/random/char_set/char_set.hh>
#include <util/random/random.hh>

#include <liblog/liblog.hh>

namespace Fred {
namespace Akm {
namespace Database {
namespace Impl {

namespace {

std::vector<LibFred::DnsKey> to_libfred_dnskeys(const Keyset& _keyset)
{
    std::vector<LibFred::DnsKey> libfred_dnskeys;
    for (const auto& dnskey : _keyset)
    {
        libfred_dnskeys.push_back(LibFred::DnsKey(
                *dnskey.flags,
                *dnskey.proto,
                *dnskey.alg,
                *dnskey.public_key));
    }
    return libfred_dnskeys;

}

struct KeysetHandleException : std::runtime_error
{
    using std::runtime_error::runtime_error;
};

std::string generate_keyset_handle(const LibFred::OperationContext& _ctx, const KeysetMetadata& _keyset_metadata)
{
    static constexpr int max_retries = 10;
    for (int i = 0; i < max_retries; ++i)
    {
        const auto handle = _keyset_metadata.handle_prefix +
            Random::Generator().get_seq(
                Random::CharSet::capital_letters_and_digits(),
                _keyset_metadata.handle_suffix_size);

        if (LibFred::TestHandleOf<LibFred::Object_Type::keyset>(handle).is_invalid_handle(_ctx))
        {
            throw KeysetHandleException{"handle is invalid"};
        }

        try
        {
            const LibFred::InfoKeysetData info_keyset_data =
                    LibFred::InfoKeysetByHandle(handle).exec(_ctx, "UTC").info_keyset_data;
        }
        catch (const LibFred::InfoKeysetByHandle::Exception& e)
        {
            if (e.is_set_unknown_handle())
            {
                return handle;
            }
            throw KeysetHandleException{"unknown error"};
        }
    }
    throw KeysetHandleException{"all generated handles were taken"};
}

std::string make_exception_message(const LibFred::CreateKeyset::Exception& e)
{
    if (e.is_set_vector_of_unknown_technical_contact_handle())
    {
        throw ConfigurationError{"unknown technical contact"};
    }
    if (e.is_set_vector_of_already_set_technical_contact_handle())
    {
        return "duplicate technical contact handle";
    }
    if (e.is_set_vector_of_already_set_dns_key())
    {
        return "duplicate dnskey";
    }
    return "unknown error";
}

std::string make_exception_message(const LibFred::UpdateDomain::Exception& e)
{
    if (e.is_set_unknown_domain_fqdn())
    {
        return "unknown domain fqdn";
    }
    if (e.is_set_unknown_keyset_handle())
    {
        return "uknown keyset handle";
    }
    return "unknown error";
}

} // namespace Fred::Akm::Database::Impl::{anonynous}

KeysetReference add_keyset(
        const LibFred::OperationContext& _ctx,
        const LibFred::InfoDomainData& _info_domain_data,
        const KeysetMetadata& _keyset_metadata,
        const Keyset& _keyset,
        boost::optional<Logger::LogEntryNumericalId> _log_entry_id)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, _info_domain_data.fqdn);
    try
    {
        const auto keyset_handle = generate_keyset_handle(_ctx, _keyset_metadata);
        const auto libfred_dnskeys = to_libfred_dnskeys(_keyset);
        const auto make_log_entry_id = [&_log_entry_id]() -> Optional<unsigned long long>
        {
            if (_log_entry_id == boost::none)
            {
                return {};
            }
            return **_log_entry_id;
        };

        const LibFred::CreateKeyset::Result create_keyset_result =
                LibFred::CreateKeyset(
                        keyset_handle,
                        _keyset_metadata.registrar,
                        Optional<std::string>(), // authinfopw
                        libfred_dnskeys,
                        Util::vector_of<std::string>(_keyset_metadata.tech_contact))
                        .exec(_ctx, make_log_entry_id(), "UTC");

        LIBLOG_DEBUG("creation_time: {} object_id: {} history_id: {}",
                        boost::to_string(create_keyset_result.creation_time),
                        create_keyset_result.create_object_result.object_id,
                        create_keyset_result.create_object_result.history_id);

        const unsigned long long domain_new_history_id = [&]()
        {
            LibFred::UpdateDomain update{_info_domain_data.fqdn, _keyset_metadata.registrar};
            update.set_keyset(keyset_handle);
            if (_log_entry_id != boost::none)
            {
                update.set_logd_request_id(**_log_entry_id);
            }
            return update.exec(_ctx);
        }();

        LIBLOG_DEBUG("domain_new_history_id: {}", domain_new_history_id);
        LibFred::Poll::CreateUpdateObjectPollMessage().exec(_ctx, domain_new_history_id);
        LIBLOG_INFO("keyset added successfully");

        return KeysetReference {
                KeysetId{create_keyset_result.create_object_result.object_id},
                KeysetHandle{keyset_handle}};
    }
    catch (const KeysetHandleException& e)
    {
        LIBLOG_ERROR("failed to create keyset handle: {}", e.what());
        throw AddKeysetFailed{e.what()};
    }
    catch (const LibFred::CreateKeyset::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw ConfigurationError{"unknown registrar handle"};
        }
        LIBLOG_ERROR("failed to create keyset: {}", make_exception_message(e));
        throw AddKeysetFailed{e.what()};
    }
    catch (const LibFred::UpdateDomain::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw ConfigurationError{"unknown registrar handle"};
        }
        LIBLOG_ERROR("failed to update domain: {}", make_exception_message(e));
        throw AddKeysetFailed{e.what()};
    }
    catch (const LibFred::OperationException& e)
    {
        LIBLOG_ERROR("failed to create \"update object poll\" message: {}", e.what());
        throw AddKeysetFailed{e.what()};
    }
    catch (const LibFred::InternalError& e)
    {
        LIBLOG_ERROR("internal error at create \"update object poll\" message: {}", e.what());
        throw AddKeysetFailed{e.what()};
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unspecified error occurred : {}", e.what());
        throw AddKeysetFailed{e.what()};
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
        throw;
    }
}

} // namespace Fred::Akm::Database::Impl
} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred
