/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/extended_info_domain_data.hh"

#include <liblog/liblog.hh>

namespace Fred {
namespace Akm {
namespace Database {

namespace {

bool is_automatically_managed_keyset(
        const KeysetMetadata& _keyset_metadata,
        const LibFred::InfoKeysetData& _info_keyset_data)
{
    if (!boost::starts_with(_info_keyset_data.handle, _keyset_metadata.handle_prefix))
    {
        return false;
    }
    return _info_keyset_data.sponsoring_registrar_handle == _keyset_metadata.registrar;
}

} // namespace Fred::Akm::Database::{anonymous}

} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred

using namespace Fred::Akm::Database;

boost::optional<ExtendedInfoDomainData> Fred::Akm::Database::get_domain_info_from_fqdn(
        const LibFred::OperationContext& _ctx,
        const KeysetMetadata& _keyset_metadata,
        const DomainFqdn& _domain_fqdn)
{
    try
    {
        const auto info_domain_data = LibFred::InfoDomainByFqdn(*_domain_fqdn)
                .exec(_ctx)
                .info_domain_data;
        if (info_domain_data.keyset.isnull())
        {
            return ExtendedInfoDomainData{info_domain_data, ScanType::insecure, boost::none};
        }

        const auto info_keyset_data = LibFred::InfoKeysetById(info_domain_data.keyset.get_value().id)
                .exec(_ctx, "UTC")
                .info_keyset_data;

        const bool domain_has_automatically_managed_keyset =
                is_automatically_managed_keyset(_keyset_metadata, info_keyset_data);
        if (domain_has_automatically_managed_keyset)
        {
            return ExtendedInfoDomainData{info_domain_data, ScanType::secure_auto, info_keyset_data};
        }
        else
        {
            return ExtendedInfoDomainData{info_domain_data, ScanType::secure, info_keyset_data};
        }
    }
    catch (const LibFred::InfoDomainByFqdn::Exception& e)
    {
        if (e.is_set_unknown_fqdn())
        {
            LIBLOG_DEBUG("domain with given fqdn was not found.");
            return boost::none;
        }
        LIBLOG_DEBUG("an error occured for domain with given fqdn: {}", e.what());
        return boost::none;
    }
}
