/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/update_keyset.hh"
#include "src/exception.hh"

#include <libfred/db_settings.hh>
#include <libfred/object/object_states_info.hh>
#include <libfred/object_state/get_object_states.hh>
#include <libfred/object_state/lock_object_state_request_lock.hh>
#include <libfred/object_state/perform_object_state_request.hh>
#include <libfred/registrable_object/keyset/update_keyset.hh>

#include <liblog/liblog.hh>

#include <algorithm>
#include <iterator>

namespace Fred {
namespace Akm {
namespace Database {
namespace Impl {

namespace {

bool are_keysets_equal(const std::vector<LibFred::DnsKey>& libfred_lhs, const Keyset& rhs)
{
    Keyset lhs;
    std::transform(cbegin(libfred_lhs), cend(libfred_lhs), std::inserter(lhs, end(lhs)), [](auto&& libfred_key)
    {
        return Dnskey{
                DnskeyFlags{libfred_key.get_flags()},
                DnskeyProto{libfred_key.get_protocol()},
                DnskeyAlg{libfred_key.get_alg()},
                DnskeyPublicKey{libfred_key.get_key()}};
    });
    return lhs == rhs;
}

bool is_keyset_shared(const LibFred::OperationContext& _ctx, unsigned long long _keyset_id)
{
    // clang-format off
    const std::string sql =
            "SELECT "
              "FROM domain "
             "WHERE keyset = $1::BIGINT "
             "LIMIT 2";
    // clang-format on
    const ::Database::Result db_result =
            _ctx.get_conn().exec_params(
                    sql,
                    ::Database::query_param_list(_keyset_id));
    return 1 < db_result.size();
}

const char* make_exception_message(const LibFred::UpdateKeyset::Exception& e)
{
    if (e.is_set_unknown_keyset_handle())
    {
        return "unknown keyset handle";
    }
    if (e.is_set_vector_of_already_set_dns_key())
    {
        return "duplicate dns key";
    }
    if (e.is_set_vector_of_unassigned_dns_key())
    {
        return "unassigned dns key";
    }
    return "unknown error";
}

} // namespace Fred::Akm::Database::Impl::{anonymous}

KeysetReference update_keyset(
        const LibFred::OperationContext& _ctx,
        const LibFred::InfoDomainData& _info_domain_data,
        const LibFred::InfoKeysetData& _info_keyset_data,
        const KeysetMetadata& _keyset_metadata,
        const Keyset& _keyset,
        boost::optional<Logger::LogEntryNumericalId> _log_entry_id)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, _info_domain_data.fqdn);
    try
    {
        if (are_keysets_equal(_info_keyset_data.dns_keys, _keyset))
        {
            LIBLOG_DEBUG("new keyset same as current keyset, nothing to do");
            throw UpdateKeysetFailed{"keyset already present"};
        }

        LibFred::LockObjectStateRequestLock(_info_keyset_data.id).exec(_ctx);
        // process object state requests
        LibFred::PerformObjectStateRequest(_info_keyset_data.id).exec(_ctx);
        const LibFred::ObjectStatesInfo keyset_states(LibFred::GetObjectStates(_info_keyset_data.id).exec(_ctx));

        if (keyset_states.presents(LibFred::Object_State::server_update_prohibited) ||
            keyset_states.presents(LibFred::Object_State::delete_candidate))
        {
            LIBLOG_WARNING("keyset state prohibits action");
            throw UpdateKeysetFailed{"keyset state prohibits action"};
        }

        if (is_keyset_shared(_ctx, _info_keyset_data.id))
        {
            LIBLOG_WARNING("automatically managed keyset referenced by multiple domains. "
                    "Cannot update keyset for security reasons. MUST BE fixed!");
            throw UpdateKeysetFailed{"automatically managed keyset referenced by multiple domains"};
        }

        LibFred::UpdateKeyset update_keyset{
                _info_keyset_data.handle,
                _keyset_metadata.registrar};

        // drop existing keys
        for (const auto& dns_key : _info_keyset_data.dns_keys)
        {
            update_keyset.rem_dns_key(dns_key);
        }
        // add new keys
        for (const auto& dnskey : _keyset)
        {
            update_keyset.add_dns_key(LibFred::DnsKey(
                    *dnskey.flags,
                    *dnskey.proto,
                    *dnskey.alg,
                    *dnskey.public_key));
        }
        if (_log_entry_id != boost::none)
        {
            update_keyset.set_logd_request_id(**_log_entry_id);
        }
        const unsigned long long keyset_new_history_id = update_keyset.exec(_ctx);

        LIBLOG_DEBUG("keyset_new_history_id: {}", keyset_new_history_id);
        return KeysetReference{
                KeysetId{_info_domain_data.keyset.get_value().id},
                KeysetHandle{_info_domain_data.keyset.get_value().handle}};
    }
    catch (const UpdateKeysetFailed&)
    {
        throw;
    }
    catch (const LibFred::UpdateKeyset::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw ConfigurationError{"unknown registrar handle"};
        }
        LIBLOG_ERROR("failed to update keyset: {}", make_exception_message(e));
        throw UpdateKeysetFailed{e.what()};
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unspecified error occurred : {}", e.what());
        throw UpdateKeysetFailed{e.what()};
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
        throw;
    }
}

} // namespace Fred::Akm::Database::Impl
} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred
