/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/add_keyset.hh"
#include "src/database/impl/extended_info_domain_data.hh"
#include "src/database/impl/is_domain_ok_for_update.hh"
#include "src/database/impl/is_keyset_ok_for_update.hh"
#include "src/database/impl/turn_off_dnssec.hh"
#include "src/database/impl/update_keyset.hh"
#include "src/database/process_acceptable_keysets.hh"
#include "src/exception.hh"
#include "src/scan_type.hh"

#include <boost/variant.hpp>

#include <libgrill/exception.hh>

#include <libfred/opcontext.hh>
#include <libfred/registrable_object/domain/info_domain.hh>

#include <liblog/liblog.hh>

#include <algorithm>
#include <cstdint>

namespace Fred {
namespace Akm {
namespace Database {

namespace {

enum class Success
{
    no,
    yes
};

struct UpdateStatus
{
    struct DomainNotFound {};

    struct NotReadyToUpdate
    {
        ScanType scan_type;
    };

    struct ReadyToUpdate
    {
        ScanType scan_type;
        Success result;
    };

    using Type = boost::variant<DomainNotFound, NotReadyToUpdate, ReadyToUpdate>;
};

void akm_turn_off(
        const LibFred::OperationContextCreator& ctx,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>> log_entry)
{
    if (log_entry == boost::none)
    {
        Impl::turn_off_dnssec(ctx, domain_info, keyset_metadata, boost::none);
        return;
    }
    try
    {
        const auto keyset_reference = Impl::turn_off_dnssec(
                ctx,
                domain_info,
                keyset_metadata,
                log_entry->get_id());
        log_entry->close_on_success(keyset_reference);
    }
    catch (...)
    {
        log_entry->close_on_failure();
        throw;
    }
}

void akm_rollover(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const LibFred::InfoKeysetData& keyset_info,
        const KeysetMetadata& keyset_metadata,
        boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_rollover>> log_entry)
{
    if (log_entry == boost::none)
    {
        Impl::update_keyset(ctx, domain_info, keyset_info, keyset_metadata, keyset, boost::none);
        return;
    }
    try
    {
        const auto keyset_reference = Impl::update_keyset(
                ctx,
                domain_info,
                keyset_info,
                keyset_metadata,
                keyset,
                log_entry->get_id());
        log_entry->close_on_success(keyset_reference, keyset_info.dns_keys, keyset);
    }
    catch (...)
    {
        log_entry->close_on_failure();
        throw;
    }
}

void akm_turn_on(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>> log_entry)
{
    if (log_entry == boost::none)
    {
        Impl::add_keyset(ctx, domain_info, keyset_metadata, keyset, boost::none);
        return;
    }
    try
    {
        const auto keyset_reference = Impl::add_keyset(
                ctx,
                domain_info,
                keyset_metadata,
                keyset,
                log_entry->get_id());
        log_entry->close_on_success(keyset_reference, keyset);
    }
    catch (...)
    {
        log_entry->close_on_failure();
        throw;
    }
}

UpdateStatus::Type process_one_domain(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeyset& _acceptable_keyset,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, *_acceptable_keyset.second.domain_fqdn);
    boost::optional<ExtendedInfoDomainData> extended_domain_info;
    try
    {
        LibFred::OperationContextCreator ctx;
        LIBLOG_DEBUG("processing keyset for domain {}: uuid {}, keyset size {}",
                *_acceptable_keyset.second.domain_fqdn,
                to_string(domain_uuid(_acceptable_keyset)),
                _acceptable_keyset.second.keyset.size());
        std::for_each(
                cbegin(_acceptable_keyset.second.keyset),
                cend(_acceptable_keyset.second.keyset),
                [fqdn = *_acceptable_keyset.second.domain_fqdn](auto&& dnskey)
        {
            LIBLOG_DEBUG("domain {}: {}, {}, {}, {}", fqdn, *dnskey.flags, *dnskey.proto, *dnskey.alg, *dnskey.public_key);
        });

        extended_domain_info = get_domain_info_from_fqdn(
                ctx,
                _keyset_metadata,
                _acceptable_keyset.second.domain_fqdn);
        if (extended_domain_info == boost::none)
        {
            LIBLOG_INFO("domain not found");
            return UpdateStatus::DomainNotFound{};
        }

        const auto domain_is_ok_for_update = Impl::is_domain_ok_for_update(
                ctx,
                extended_domain_info->info_domain_data,
                domain_uuid(_acceptable_keyset),
                _keyset_metadata);
        if (!domain_is_ok_for_update)
        {
            LIBLOG_INFO("domain is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }
        const auto make_successful_result = [&]()
        {
            if (dry_run != DryRun::yes)
            {
                ctx.commit_transaction();
            }
            return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::yes};
        };
        if (Impl::is_dnssec_turn_off_requested(_acceptable_keyset.second.keyset))
        {
            if (extended_domain_info->scan_type == ScanType::insecure)
            {
                LIBLOG_INFO("turn off requested for insecure domain, nothing to do.");
                return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
            }
            const auto create_log_entry = [&](DryRun dry_run) -> boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_off>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return boost::none;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_turn_off(
                    ctx,
                    extended_domain_info->info_domain_data,
                    _keyset_metadata,
                    create_log_entry(dry_run));
            return make_successful_result();
        }
        if (!Impl::is_keyset_ok_for_update(ctx, _acceptable_keyset.second.keyset))
        {
            LIBLOG_INFO("keyset is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }

        if (extended_domain_info->scan_type == ScanType::secure_auto)
        {
            const auto create_log_entry = [&](DryRun dry_run) -> boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_rollover>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_rollover>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return boost::none;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_rollover(
                    ctx,
                    _acceptable_keyset.second.keyset,
                    extended_domain_info->info_domain_data,
                    *extended_domain_info->info_keyset_data,
                    _keyset_metadata,
                    create_log_entry(dry_run));
            return make_successful_result();
        }
        const auto create_log_entry = [&](DryRun dry_run) -> boost::optional<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>>
        {
            switch (dry_run)
            {
                case DryRun::no:
                    return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_on>(
                            DomainId{extended_domain_info->info_domain_data.id},
                            _acceptable_keyset.second.domain_fqdn);
                case DryRun::yes:
                    return boost::none;
            }
            throw std::runtime_error{"unexpected DryRun value"};
        };
        akm_turn_on(
                ctx,
                _acceptable_keyset.second.keyset,
                extended_domain_info->info_domain_data,
                _keyset_metadata,
                create_log_entry(dry_run));
        return make_successful_result();
    }
    catch (const LibGrill::GrpcException& e) // cannot be caught by std::exception; TODO: fix libgrill
    {
        LIBLOG_ERROR("LibGrill::GrpcException caught: {}", e.error_message());
        throw ConfigurationError{"logger misconfigured"};
    }
    catch (const ConfigurationError& e)
    {
        LIBLOG_ERROR("ConfigurationError exception: {}", e.what());
        throw;
    }
    catch (const Impl::UpdateKeysetFailed& e)
    {
        LIBLOG_INFO("UpdateKeysetFailed exception: {}", e.what());
    }
    catch (const Impl::OperationFailed& e)
    {
        LIBLOG_INFO("OperationFailed exception: {}", e.what());
    }
    catch (const std::exception& e)
    {
        LIBLOG_INFO("std::exception: {}", e.what());
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
    }
    if (extended_domain_info != boost::none)
    {
        return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::no};
    }
    return UpdateStatus::DomainNotFound{};
}

class Statistics
{
public:
    void add(const UpdateStatus::Type& update_status)
    {
        boost::apply_visitor(*this, update_status);
    }
    void print()
    {
        std::for_each(cbegin(type_to_data_), cend(type_to_data_), [](auto&& type_data_pair)
        {
            const auto& type = type_data_pair.first;
            const auto& data = type_data_pair.second;
            LIBLOG_INFO("======== update stats {} candidates =======", to_string(type));
            LIBLOG_INFO("   domains loaded:          {}", data.total());
            LIBLOG_INFO("    ├─ ok_for_update:       {}", data.ok_for_update());
            LIBLOG_INFO("    │   ├─ updated:         {}", data.ok_for_update_success);
            LIBLOG_INFO("    │   └─ failed:          {}", data.ok_for_update_failure);
            LIBLOG_INFO("    └─ ko_for_update:       {}", data.not_ok_for_update);
        });
        if (domains_not_found_ != 0)
        {
            LIBLOG_WARNING("not found domains:       {}", domains_not_found_);
        }
    }
    void operator()(const UpdateStatus::DomainNotFound&)
    {
        ++domains_not_found_;
    }
    void operator()(const UpdateStatus::NotReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_not_ready();
    }
    void operator()(const UpdateStatus::ReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_ready(status.result);
    }
private:
    struct Data
    {
        std::int64_t ok_for_update_success = 0;
        std::int64_t ok_for_update_failure = 0;
        std::int64_t not_ok_for_update = 0;
        constexpr std::int64_t ok_for_update() const noexcept
        {
            return ok_for_update_success + ok_for_update_failure;
        }
        constexpr std::int64_t total() const noexcept
        {
            return ok_for_update() + not_ok_for_update;
        }
        constexpr void add_not_ready()
        {
            ++not_ok_for_update;
        }
        constexpr void add_ready(Success result)
        {
            switch (result)
            {
                case Success::no:
                    ++ok_for_update_failure;
                    return;
                case Success::yes:
                    ++ok_for_update_success;
                    return;
            }
        }
    };
    std::int64_t domains_not_found_ = 0;
    std::map<ScanType, Data> type_to_data_ = {
            {ScanType::insecure, {}},
            {ScanType::secure, {}},
            {ScanType::secure_auto, {}}};
};

} // namespace Fred::Akm::Database::{anonymous}

void process_acceptable_keysets(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeysets& _acceptable_keysets,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    Statistics statistics;
    std::for_each(cbegin(_acceptable_keysets), cend(_acceptable_keysets), [&](auto&& acceptable_keyset)
    {
        const auto update_status = process_one_domain(
                _keyset_metadata,
                acceptable_keyset,
                _log_entry_factory,
                dry_run);
        statistics.add(update_status);
    });
    statistics.print();
}

} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred
