/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/logger/log_entry.hh"
#include "src/logger/impl/operation_result.hh"
#include "src/logger/impl/to_log_entry_property.hh"
#include "src/logger/impl/to_object_reference.hh"

#include <libgrill/logger/close_log_entry.hh>

namespace Fred {
namespace Akm {
namespace Logger {

template <LogEntryType T>
LogEntry<T>::LogEntry(
        LibGrill::Connection<LibGrill::Service::Logger>& _connection,
        const LibGrill::Logger::LogEntryId& _log_entry_id,
        const DomainFqdn& _domain_fqdn)
    : connection_(_connection), log_entry_id_(_log_entry_id), domain_fqdn_(_domain_fqdn)
{
}

template <LogEntryType T>
LogEntryNumericalId LogEntry<T>::get_id() const
{
    return log_entry_id_.get_numerical_id();
}

namespace {

using namespace ::LibGrill::Logger;

CloseLogEntryRequest close_log_entry_on_success_request(
        const LogEntryId& _log_entry_id,
        const DomainFqdn& _domain_fqdn,
        const KeysetReference& _keyset)
{
    CloseLogEntryRequest request{
            _log_entry_id,
            to_libgrill_operation_result_name<OperationResult::success>()};
    request.add_object_reference(_keyset.id);
    request.add_property(_domain_fqdn);
    request.add_property(_keyset.handle);
    return request;
}

void add_new_keyset(CloseLogEntryRequest& _request, const Keyset& _keyset)
{
    for(const auto& key : _keyset)
    {
        _request.add_property(key);
    }
}

void add_old_keyset(CloseLogEntryRequest& _request, const std::vector<LibFred::DnsKey>& _keyset)
{
    for(const auto& key : _keyset)
    {
        _request.add_property(key);
    }
}

} // namespace Fred::Akm::Logger::{anonymous}

template<>
template<>
void LogEntry<LogEntryType::akm_turn_off>::close_on_success(const KeysetReference& _keyset)
{
    auto request = close_log_entry_on_success_request(log_entry_id_, domain_fqdn_, _keyset);
    ::LibGrill::Logger::close_log_entry(connection_, request);
}

template<>
template<>
void LogEntry<LogEntryType::akm_turn_on>::close_on_success(
        const KeysetReference& _keyset_reference,
        const Keyset& _new_keyset)
{
    auto request = close_log_entry_on_success_request(log_entry_id_, domain_fqdn_, _keyset_reference);
    add_new_keyset(request, _new_keyset);
    ::LibGrill::Logger::close_log_entry(connection_, request);
}

template<>
template<>
void LogEntry<LogEntryType::akm_rollover>::close_on_success(
        const KeysetReference& _keyset_reference,
        const std::vector<LibFred::DnsKey>& _old_keyset,
        const Keyset& _new_keyset)
{
    auto request = close_log_entry_on_success_request(log_entry_id_, domain_fqdn_, _keyset_reference);
    add_new_keyset(request, _new_keyset);
    add_old_keyset(request, _old_keyset);
    ::LibGrill::Logger::close_log_entry(connection_, request);
}

template<LogEntryType T>
void LogEntry<T>::close_on_failure()
{
    CloseLogEntryRequest request(
            log_entry_id_,
            to_libgrill_operation_result_name<OperationResult::fail>());
    ::LibGrill::Logger::close_log_entry(connection_, request);
}

template class LogEntry<LogEntryType::akm_turn_off>;
template class LogEntry<LogEntryType::akm_turn_on>;
template class LogEntry<LogEntryType::akm_rollover>;

} // namespace Fred::Akm::Logger
} // namespace Fred::Akm
} // namespace Fred
